require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

RSpec.shared_examples 'entity reader only' do
  it 'returns status code 200' do
    authenticated_header(request, reader)
    subject
    expect(response).to have_http_status 200
  end
  it 'returns unauthorized error' do
    authenticated_header(request, user)
    subject
    expect(response).to have_http_status 401
  end
end

describe Hermes::Api::Dmo::HousingRequestEntityFormLevelsController, :type => :api do
  routes { Hermes::Engine.routes }
  include Devise::Test::ControllerHelpers

  check_mutate_params(
    model: Hermes::HousingRequestEntityFormLevel, 
    mutate_params: Hermes::Api::Dmo::HousingRequestEntityFormLevelsController.mutate_params, 
    factory: :entity_form_level,
    other_factory:[
      {factory: :entity, id: 'entity1'},
    ],
    embedded_params: {
      entity_id: 'entity1',
    },
    check_put: false,
    not_checked_keys: [:entity_id]
  )

  let!(:responsible) { create(:user, roles: ['Administrator']) }
  let!(:entreprise) { create(:user, auth_roles: ['entreprise'], roles: []) }
  let!(:demandeur) { create(:user, auth_roles: ['demandeur']) }
  let!(:user) { create :user }

  let(:reader) { [responsible, entreprise, demandeur].sample }

  let!(:entity) { create :entity, readers: [demandeur.id.to_s], writers: [entreprise.id.to_s], siret: '12345678901234' }
  let(:form_level) { create :entity_form_level, entity_id: entity.id, consent_fields: ['firstname', 'lastname'], writers: [entreprise.id.to_s] }
  let(:params) do
    { attributes: { consent_fields: ['field_1', 'field_2'], description: 'Constement niveau2', entity_id: entity.id } }
  end

  describe 'GET index' do
    let!(:form_levels) { create_list :entity_form_level, 5, entity_id: entity.id }
    # let!(:user_right) { create(:user_right, user_id: entreprise.id, entity_id: entity.id, right: 'administrator', type: 'entity')}

    before do
      authenticated_header(request, entreprise)
    end
    
    it 'with entity_siret' do
      get 'index', entity_siret: entity.siret
      expect(response).to have_http_status 200
      expect(json['data'].length).to eq 5
      expect(json['meta']['pagination']).to be_present
      expect(json['meta']['pagination']['total_objects']).to eq 5
    end

    it 'with entity_id' do
      get 'index', entity_id: entity.id
      expect(response).to have_http_status 200
      expect(json['data'].length).to eq 5
      expect(json['meta']['pagination']).to be_present
      expect(json['meta']['pagination']['total_objects']).to eq 5
    end

    %w[
      cre
      entreprise
      bailleur
      rdsl
    ].each do |type|
      context "as #{type}" do
        let!(:user_right) { create(:user_right, user_id: user.id, entity_id: entity.id, right: 'administrator', type: 'entity')}

        before do 
          user.update(auth_roles: [type])
          authenticated_header(request, user)
        end 
        context 'with user_right' do 
          before do 
            get 'index', entity_siret: entity.siret
          end 
          it { expect(response).to have_http_status(200) }
        end
        context 'without user_right' do 
          before do
            user_right.destroy
            get 'index', entity_siret: entity.siret
          end
          it { expect(response).to have_http_status(401) }
        end
      end 
    end 

    context 'as superadministrateur' do 
      let!(:superadministrateur) { create(:user, auth_roles: ['superadministrateur'])}
      before  do 
        authenticated_header(request, superadministrateur)
      end 

      it 'has access' do
        get 'index', entity_siret: entity.siret
        expect(response).to have_http_status(200)
      end
    end 

    context 'when demandeur' do 
      let!(:hr) { create(:active_request)}

      before :each do 
        authenticated_header(request, demandeur)
      end 

      it 'no access when only demandeur with no active request' do 
        get 'index', entity_siret: entity.siret
        expect(response).to have_http_status(401)
      end 
  
      it 'has access when demandeur with active request' do 
        hr.set(is_active: true, entity_id: entity.id, user_id: demandeur.id)
        get 'index', entity_siret: entity.siret
        expect(response).to have_http_status(200)
      end 
    end 


  end

  describe 'GET show' do
    subject { get 'show', {id: form_level.id } }

    it_behaves_like 'entity reader only'

    it 'returns a valid json' do
      authenticated_header(request, reader)
      subject
      expect(json['data']['id']).to eq form_level.id.to_s
      expect(json['data']['attributes']['consent_fields']).to eq form_level.consent_fields
    end
  end

  describe 'POST create' do
    subject { post 'create', { data: params } }

    context 'when entreprise' do
      before { authenticated_header(request, entreprise) }
      it 'succeeds' do
        subject
        expect(response).to have_http_status 201
      end
      it 'creates an form_level for entity' do
        expect { subject }.to change { entity.form_levels.count }.by 1
      end
      it 'returns a valid json' do
        subject
        expect(json['data']['attributes']['consent_fields']).to eq ['field_1', 'field_2']
        expect(json['data']['attributes']['description']).to eq 'Constement niveau2'
      end
    end

    context 'when other entity reader' do
      before { authenticated_header(request, demandeur) }
      it 'returns unauthorized error' do
        subject
        expect(response).to have_http_status 401
      end
    end

    context 'when not entity reader' do
      before { authenticated_header(request, user) }
      it 'returns unauthorized error' do
        subject
        expect(response).to have_http_status 401
      end
    end
  end

  describe 'PATCH update' do
    subject { patch 'update', { id: form_level.id, data: params } }

    context 'when form_level writer' do
      before { authenticated_header(request, entreprise) }
      it 'succeeds' do
        subject
        expect(response).to have_http_status 200
      end
      it 'updates the form_level' do
        expect { subject }.to change { form_level.reload.consent_fields }.from(['firstname', 'lastname']).to ['field_1', 'field_2']
      end
      it 'returns a valid json' do
        subject
        expect(json['data']['id']).to eq form_level.id.to_s
        expect(json['data']['attributes']['description']).to eq 'Constement niveau2'
      end
    end

    context 'when not form_level writer' do
      before do
        form_level.update!(writers: [])
        authenticated_header(request, entreprise)
      end
      it 'returns not_writers error' do
        subject
        expect(response).to have_http_status 422
      end
    end

    context 'when not authorized' do
      before { authenticated_header(request, user) }
      it 'returns unauthorized error' do
        subject
        expect(response).to have_http_status 401
      end
    end
  end

  describe 'DELETE destroy' do
    let!(:owned_form_level) { create :entity_form_level, entity_id: entity.id, user_id: responsible.id }

    before do
      create(:user_right, user_id: responsible.id, entity_id: entity.id,  type: 'entity', right: "administrator")
      authenticated_header(request, responsible)
    end

    context 'when form_level owned' do
      subject { delete 'destroy', { id: owned_form_level.id } }
      it 'succeeds' do
        subject
        expect(response).to have_http_status 204
      end

      it 'destroys the form_level' do
        expect { subject }.to change { entity.form_levels.count }.by -1
      end
    end

    context 'when form_level not owned' do
      subject { delete 'destroy', { entity_id: entity.id, id: form_level.id } }
      before { authenticated_header(request, responsible) }

      it 'succeeds' do
        subject
        expect(response).to have_http_status 204
      end
    end
  end
end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::HousingRequestEntityFormLevelsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_HOUSING_REQUEST_ENTITY_FORM_LEVELS'

  
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:admin) { create(:user, roles: ['Administrator']) }
  let!(:entreprise) { create(:user, roles: ['entreprise']) }
  let!(:entity) { create :entity, user_id: entreprise.id }
  let!(:entity_id) { entity.id }


  path '/api/dmo/housing_request_entity_form_levels' do
    get 'Retrieves all housing_request_entity_form_levels' do
      tags tags
      produces 'application/json'
      parameter name: :entity_siret, in: :query, type: :string, required: false
      Hermes::Api::Dmo::HousingRequestEntityFormLevelsController.acceptable_params.sort.each do |params_name, options|
        parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
      end

      security [Bearer: {}]
      paginator

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '200', 'housing_request_entity_form_levels found' do
        let(:Authorization) { "Bearer #{token_for(entreprise)}" }
        let!(:form_levels) { create_list(:entity_form_level, 5, entity_id: entity_id) }

        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) { form_levels }
        end
      end
    end
  end

  path '/api/dmo/housing_request_entity_form_levels/{id}' do
    get 'Retrieves a single housing_request_entity_form_levels by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let(:form_level) { create(:entity_form_level, entity_id: entity_id) }

      let(:id) {form_level.id}

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '200', 'entity_form_level found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}

        include_context 'with integration test'
        include_context 'valid jsonapi ressource' do
          let(:expected_ressource) {form_level}
        end
      end
    end
  end

  path '/api/dmo/housing_request_entity_form_levels' do
    post 'Create housing_request_entity_form_level' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::HousingRequestEntityFormLevelsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )

      let(:params) do
        jsonapi_attributes(
          {
            entity_id: entity.id,
            level: 4,
          }
        )
      end

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '201', 'housing_request_entity_form_level created' do
        let(:Authorization) { "Bearer #{token_for(entreprise)}" }
        let(:params) do
          jsonapi_attributes({})
        end
        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'housing_request_entity_form_levels'
        end
      end
    end
  end

  path '/api/dmo/housing_request_entity_form_levels/{id}' do
    put 'edit housing_request_entity_form_level' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::HousingRequestEntityFormLevelsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )
      let(:form_level) { create(:entity_form_level, entity_id: entity.id) }
      let!(:user_right) { create(:user_right, user_id: admin.id, entity_id: entity.id, type: 'entity', right: 'administrator') }
      let(:id) { form_level.id }
      let(:params) do
        jsonapi_attributes(
          {
            entity_id: entity.id,
            level: 4,
          }
        )
      end

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '200', 'housing_request_entity_form_level updated' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)
          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to eq form_level.id.to_s
          expect(json['data']['type']).to eq 'housing_request_entity_form_levels'
        end
      end
    end
  end

  path '/api/dmo/housing_request_entity_form_levels/{id}' do
    delete 'Delete a single housing_request_entity_form_level by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let(:form_level) { create(:entity_form_level, entity_id: entity_id, writers: [entreprise.id]) }
      let(:id) { form_level.id }

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '204', 'housing_request_entity_form_level deleted' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:id) { create(:entity_form_level, entity_id: entity_id, user_id: admin.id, writers: [admin.id]).id }
        let!(:user_right) { create(:user_right, user_id: admin.id, entity_id: entity.id, type: 'entity', right: "administrator") }
        run_test!
      end
    end
  end
end