require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

describe Hermes::Api::Dmo::ManagementAreasController, :type => :api do
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers

  check_mutate_params(
    model: Hermes::ManagementArea, 
    mutate_params: Hermes::Api::Dmo::ManagementAreasController.mutate_params, 
    factory: :management_area,
    min_params: {
      name: 'test',
    }
  )

  let!(:responsible) { create(:user, roles: ['Administrator']) }
  let!(:user) { create(:user) }
  let!(:lessor) { create(:lessor, siret: '00000000000000', readers: [user.id.to_s]) }
  let!(:lessor2) { create(:lessor, siret: '00000000000000', readers: [responsible.id.to_s]) }
  
  let!(:management_area1) { create(:management_area, name: 'name1', lessor_id: lessor2.id.to_s) }
  let!(:management_area2) { create(:management_area, name: 'name2', lessor_id: lessor2.id.to_s) }
  let!(:management_area3) { create(:management_area, name: 'test3', lessor_id: lessor2.id.to_s) }
  let!(:management_area4) { create(:management_area, name: 'test4', lessor_id: lessor.id.to_s) }
  
  describe 'POST /management_areas' do
    describe 'Create as admin' do
      before :each do
        authenticated_header(request, responsible)
        post 'create', {
            data: {
                attributes: {
                    name: 'test'
                }
            }
        }
      end
      it 'returns status code 201' do

        expect(response).to have_http_status(201)
        expect(json['data']['attributes']['code']).to eq('ALSSG-4')
      end
    end

    describe 'Create as user without lessor_id' do
      before :each do
        authenticated_header(request, user)
        post 'create', {
            data: {
                attributes: {
                    name: 'test'
                }
            }
        }
      end
      it 'returns error when no lessor_id' do
        expect(response).to have_http_status(422)
      end
    end

    describe 'Create as user with lessor_id' do
      before :each do
        authenticated_header(request, user)
        post 'create', {
            data: {
                attributes: {
                    name: 'test',
                    lessor_id: lessor.id.to_s
                }
            }
        }
      end
      it 'returns 201' do
        expect(response).to have_http_status(201)
        expect(json['data']['attributes']['code']).to eq('ALSSG-4')
      end
    end
  end

  describe 'GET /management_areas' do
    describe 'as admin' do 
      before :each do
        authenticated_header(request, responsible)
        get 'index'
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(4)
      end
    end
    describe 'as user' do 
      before :each do
        authenticated_header(request, user)
        get 'index'
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(1)
      end
    end

    describe 'regex on name' do 
      before :each do
        authenticated_header(request, responsible)
        get 'index', {name: {"$regex": "test" }}
      end
      it 'returns a valid json' do
        expect(response).to have_http_status(200)
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(2)
      end
    end

  end

  describe 'GET /management_areas/:id' do
    describe 'verif it work' do
      before :each do
        authenticated_header(request, user)
        get 'show', {id: management_area4.id}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
      end
    end

    describe 'verif it work not work with a non authorized user' do

      before :each do
        authenticated_header(request, user)
        get 'show', {id: management_area1.id}
      end
      it 'returns status code 404' do
        expect(response).to have_http_status(404)
      end
    end
  end


  describe 'DELETE /management_areas/:id' do
    before :each do
      authenticated_header(request, responsible)

      delete 'destroy', {
          id: management_area1.id.to_s
      }
    end
    it 'returns status code 204' do
      expect(response).to have_http_status(204)
    end
  end

end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::ManagementAreasController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_MANAGEMENT_AREAS'
  
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:admin) { create(:user, roles: ['Administrator']) }

  path '/api/dmo/management_areas' do
    get 'Retrieves all management_areas' do
      tags tags
      produces 'application/json'

      Hermes::Api::Dmo::ManagementAreasController.acceptable_params.sort.each do |params_name, options|
        parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
      end

      security [Bearer: {}]
      paginator

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '200', 'management_areas found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let!(:management_areas) { create_list(:management_area, 5, name: 'test') }

        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) { management_areas }
        end
      end
    end
  end

  path '/api/dmo/management_areas' do
    post 'Create management_area' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::ManagementAreasController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )
      let(:params) do
        jsonapi_attributes(
          {  name: 'test' }
        )
      end

      response '201', 'management_area created' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'management_areas'
        end
      end
    end
  end

  path '/api/dmo/management_areas/{id}' do
    get 'Retrieves a single management_area by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        let(:id) { 1 }

        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }

        run_test!
      end 
      
      response '200', 'management_area found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:management_area) {create(:management_area, name: 'test')}
        let(:id) {management_area.id}

        include_context 'with integration test'
        include_context 'valid jsonapi ressource' do
          let(:expected_ressource) {management_area}
        end
      end
    end
  end

  path '/api/dmo/management_areas/{id}' do
    put 'Update management_area' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::ManagementAreasController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter name: :id, in: :path, type: :string
      let(:id) { create(:management_area, name: 'test', user: admin).id }
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )
      let(:params) do
        jsonapi_attributes(
          {  name: 'test' }
        )
      end

      response '200', 'management_area updated' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)
          
          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'management_areas'
        end
      end
    end
  end

  path '/api/dmo/management_areas/{id}' do
    delete 'Delete a single management_area by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]

      response '204', 'management_area deleted' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { create(:management_area, name: 'test').id }
        run_test!
      end
    end
  end

end
