require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::PicturesController, :type => :api do
  
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers
  
  let!(:responsible) {create(:user, roles: ['Administrator'])}
  let!(:user) {create(:user)}

  let!(:offer) {create(:housing_offer)}
  
  describe 'POST /housing_offers/{:housing_offer_id}/pictures' do
    let!(:cat_picture) {Rack::Test::UploadedFile.new(File.join(Rails.root, 'packages', 'hermes', 'spec', 'fixtures', 'files', 'cat1.jpg'), 'image/jpeg') }
    let!(:pdf_as_image) {Rack::Test::UploadedFile.new(File.join(Rails.root, 'packages', 'hermes', 'spec', 'fixtures', 'files', 'pdf_as_image.jpg'), 'image/jpeg') }
    let!(:params) { { housing_offer_id: offer.id.to_s, file: cat_picture} }

    describe 'create' do
      before :each do
        authenticated_header(request, responsible)
      end 

      it 'works' do
        post 'create', params
        expect(response).to have_http_status(201)
        expect(offer.reload.ordered_picture_ids.length).to eq 1
      end

      it 'when housing_offer_not_found' do 
        post 'create', params.merge(housing_offer_id: 'test')
        expect(response).to have_http_status(404)
        expect(response.body).to include('housing_offer_not_found')
      end 

      it 'has user lessor with access' do
        lessor = create(:lessor, siren: '123456789')
        user_lessor = create(:user, auth_roles: ['lessor'])
        user_right = create(:user_right, user_id: user_lessor.id, type: 'lessor', lessor_id: lessor.id.to_s, right: 'administrator')
        offer.update(lessor_siren: lessor.siren)
        authenticated_header(request, user_lessor)
        post 'create', params
        expect(response).to have_http_status(201)
        expect(offer.reload.ordered_picture_ids.length).to eq 1
      end

      it 'when no access' do
        authenticated_header(request, user)
        post 'create', params
        expect(response).to have_http_status(401)
        expect(response.body).to include('no_access')
      end

      it 'when no file' do 
        post 'create', params.except(:file)
        expect(response).to have_http_status(422)
        expect(response.body).to include('no_file_in_the_request')
      end 

      it 'when number of picture execeded' do
        20.times do |i|
          create(:picture, housing_offer_id: offer.id.to_s, user_id: responsible.id)
        end 
        authenticated_header(request, responsible)
        post 'create', params
        expect(response).to have_http_status(422)
        expect(response.body).to include('number_of_pictures_exeded')
      end

      it 'when unsupported media type' do 
        post 'create', params.merge(file: pdf_as_image)
        expect(response).to have_http_status(415)
        expect(response.body).to include('not_an_image')
      end 
    end
  end

  describe 'DELETE /housing_offers/{:housing_offer_id}/pictures/{:id}' do
    let!(:picture) {create(:picture, housing_offer_id: offer.id.to_s, user_id: responsible.id)}

    before :each do
      authenticated_header(request, responsible)
    end 

    it 'works' do
      delete 'destroy', housing_offer_id: offer.id.to_s, id: picture.id.to_s
      expect(response).to have_http_status(204)
      expect(offer.reload.ordered_picture_ids.length).to eq 0
    end 

    it 'when no housing_offer found' do
      delete 'destroy', housing_offer_id: 'test', id: picture.id.to_s
      expect(response).to have_http_status(404)
      expect(response.body).to include('housing_offer_not_found')
    end

    it 'when no picture found' do
      delete 'destroy', housing_offer_id: offer.id.to_s, id: 'test'
      expect(response).to have_http_status(404)
      expect(response.body).to include('picture_not_found')
    end

    it 'when no access' do
      authenticated_header(request, user)
      delete 'destroy', housing_offer_id: offer.id.to_s, id: picture.id.to_s
      expect(response).to have_http_status(401)
      expect(response.body).to include('no_access')
    end 
  end 
end

describe Hermes::Api::Dmo::PicturesController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  TAG = 'API_DMO_PICTURES'
  
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
  let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}
  let!(:applicant) {create(:applicant, user: responsible)}
  let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
  let!(:als_rh) {Hermes::ReservationHolder.create_default}

  # Pictures functionalities
  path '/api/dmo/housing_offers/{housing_offer_id}/pictures' do
    post 'create_picture for an housing_offer' do
      tags TAG
      produces 'application/json'
      parameter name: 'Content-Type', in: :header, type: :string, required: false, description: "must be 'multipart/form-data'"
      consumes 'multipart/form-data'
      # consumes 'application/json'
      parameter name: :housing_offer_id, in: :path, type: :string
      security [Bearer: {}]
      parameter name: :file, in: :formData, type: :file, description: 'picture to upload'
      description 'Send a file with the post action to add a picture to the housing_offer. you can only add 20 pictures, use delete and then post do add others or use put to change one.<br> Edit the housing_offer[ordered_picture_ids] field. the main picture is the first of the array. <br> The main picture is the first of the array'

      response '201', 'picture created' do
        let(:file) do
          Rack::Test::UploadedFile.new([Hermes::Engine.root, %w[spec xml_bouchon minion.jpg]].flatten.join('/'), 'mime/type')
        end
        let!(:housing_offer1) {create(:housing_offer)}
        let(:'Content-Type') {'multipart/form-data'}
        let(:Authorization) {"Bearer #{token_for(responsible)}"}
        let(:housing_offer_id) {housing_offer1.id}
        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json_data = JSON.parse(response.body)['data']
          ho = Hermes::HousingOffer.find_by(id: housing_offer1.id)
          expect(json_data).to be_a Hash
          expect(json_data['id']).to_not be nil
          expect(json_data['type']).to eq 'pictures'
          expect(json_data['attributes']['housing_offer_id']).to eq(housing_offer1.id.to_s)
          expect(json_data['attributes']['thumb240']).to_not be_nil
          expect(
            Hermes::HousingOfferSerializer.new(
              Hermes::HousingOffer.find_by(id: json_data['attributes']['housing_offer_id']),
              scope: responsible,
              root: true
            ).attributes[:pictures_nb]
          ).to eq(1)
          expect(ho.ordered_picture_ids).to eq([json_data['id'].to_s])
        end
      end
    end
  end
  path '/api/dmo/housing_offers/{housing_offer_id}/pictures/{picture_id}' do
    put 'edit_picture for an housing_offer' do
      tags TAG
      parameter name: :housing_offer_id, in: :path, type: :string
      parameter name: :picture_id, in: :path, type: :string
      produces 'application/json'
      parameter name: 'Content-Type', in: :header, type: :string, required: false, description: "must be 'multipart/form-data'"
      consumes 'multipart/form-data'
      parameter name: :file, in: :formData, type: :file, description: 'picture to upload'
      security [Bearer: {}]
      description 'Send a file with the update action to edit one of the picture. if you want to reorder picture, edit the housing_offer[ordered_picture_ids] field. the main picture is the first of the array'

      response '200', 'picture edited' do
        let(:file) do
          Rack::Test::UploadedFile.new([Hermes::Engine.root, %w[spec xml_bouchon minion.jpg]].flatten.join('/'), 'mime/type')
        end
        let!(:housing_offer1) {create(:housing_offer)}
        let(:Authorization) {"Bearer #{token_for(responsible)}"}
        let(:housing_offer_id) {housing_offer1.id}
        let(:picture1) {create(:picture, housing_offer_id: housing_offer1.id, user_id: responsible.id.to_s)}
        let(:picture_id) {picture1.id.to_s}
        let(:old_assets_ids) {[picture1.thumb240_id, picture1.thumb1000_id, picture1.full_size_id]}
        let(:'Content-Type') {'multipart/form-data'}
        # include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'
        run_test!
        after do |_example|
          json_data = JSON.parse(response.body)['data']
          expect(json_data).to be_a Hash
          expect(json_data['id']).to_not be nil
          edited_pic = Hermes::Picture.find_by(id: picture1.id)
          expect(json_data['type']).to eq 'pictures'
          expect(edited_pic.thumb240_id).not_to eq(old_assets_ids[0])
          expect(edited_pic.thumb1000_id).not_to eq(old_assets_ids[1])
          expect(edited_pic.full_size_id).not_to eq(old_assets_ids[2])
        end
      end
    end
  end
  path '/api/dmo/housing_offers/{housing_offer_id}/pictures/{id}' do
    delete 'delete_picture' do
      tags TAG
      produces 'application/json'
      consumes 'application/json'
      parameter name: :housing_offer_id, in: :path, type: :string, required: true, description: 'ID of the housing offer'
      parameter name: :id, in: :path, type: :string, required: true, description: 'ID of the picture'
      security [Bearer: {}]

      response '204', 'picture deleted' do
        let(:Authorization) {"Bearer #{token_for(responsible)}"}
        let(:housing_offer_id) {create(:housing_offer).id.to_s}
        let(:id) {create(:picture, user_id: responsible.id, housing_offer_id: housing_offer_id).id.to_s}

        run_test!
        after do |_example|
          expect(Hermes::HousingOffer.find(housing_offer_id).ordered_picture_ids).to eq([])
        end
      end
    end
  end
end