require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

RSpec.describe Hermes::HousingOffer, type: :model do
  it 'has a valid factory' do
    Hermes::ReservationHolder.create_default
    expect(FactoryBot.build(:housing_offer)).to be_valid
  end

  let!(:threshold_plus) do
    create(
      :threshold,
      type: "PLUS",
      codes: ["PLU"],
      thresholds_map: {
        "Paris et communes limitrophes"=>{"thresholds_list"=>[23721, 35452, 46473, 55486, 66017, 74286], "another_person"=>8278},
        "IDF hors Paris et Communes limitrophes"=>{"thresholds_list"=>[23721, 35452, 42616, 51046, 60429, 68001], "another_person"=>7577},
        "Autres régions"=>{"thresholds_list"=>[20623, 27540, 33119, 39982, 47035, 53008], "another_person"=>5912}
      }
    )
  end

  describe 'validate save side charges effect' do
    let!(:basic_housing_offer) do
      create(
        :housing_offer,
        rental_charges: 24.5,
        parking_charges: 12.3,
        rent_amount: 78.0,
        rent_with_charges: 78.0 + 12.3 + 24.5,
        typology: 'T2',
        heating_type: 'Individuel électrique'
      )
    end

    it 'put referent' do
      expect(basic_housing_offer.total_charges).to eq(108.52)
    end
  end

  describe 'without heating charges' do
    let!(:basic_housing_offer) do
      create(
        :housing_offer,
        rental_charges: 24.5,
        parking_charges: 12.3,
        rent_amount: 78.0,
        rent_with_charges: 78.0 + 12.3 + 24.5,
        typology: 'T1',
        heating_type: 'Collectif'
      )
    end

    it 'put referent' do
      expect(basic_housing_offer.total_charges).to eq(36.8)
    end
  end

  describe 'validate save is_mad side effect' do
    let!(:housing_offer1) do
      create(
        :housing_offer,
        lessor_status: 'PLACE',
        offer_status: "EN ATTENTE COMMERCIALISATION"
      )
    end

    let!(:housing_offer2) do
      create(
        :housing_offer,
        lessor_status: 'RETIRE'
      )
    end

    let!(:housing_offer3) do
      create(
        :housing_offer,
        offer_status: 'RETIRE PAR LE BAILLEUR'
      )
    end

    let!(:housing_offer4) do
      create(
        :housing_offer,
        offer_status: 'LOGEMENT PLACE'
      )
    end

    it 'put referent' do
      expect((1..4).map { |i| send("housing_offer#{i}").is_mad }).to all(be false)
    end

    describe 'when status gets changed to mad situation' do
      before :each do
        housing_offer1.update(lessor_status: 'EN ATTENTE PUBLICATION')
      end

      it 'should has the right status' do
        expect(housing_offer1.reload.lessor_status).to eq 'EN ATTENTE PUBLICATION'
      end

      it 'should has updated correctly' do
        offer1 = housing_offer1.reload
        expect(offer1.is_mad).to be true
        expect(offer1.is_mad_updated_at).to_not eq offer1.created_at
      end

      it 'send flux 5' do 
        job = enqueued_jobs.find { |ej| ej[:job] == ExecuteFlux5TaskJob }
        expect(job[:args]).to eq([housing_offer1.id.to_s])
        expect(job[:at] > (Time.now + 30.seconds).to_time.to_i).to be true
      end 
    end
  end

  describe 'Update realized_by_typologies of linked lessor_objective' do
    let!(:linked_lessor) { create(:lessor) }

    let!(:basic_housing_offer) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'not rented',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # à supprimer une fois les statuts mis à jour
      )
    end

    let!(:linked_lessor_objective) do
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:all =>100},
      )
    end

    it 'setting up' do
      linked_lessor_objective.reload
      expect(linked_lessor_objective.target_by_typologies).to(eq({'all'=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq(['all']))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({'all'=>0}))
    end
  end

  describe 'Realized calculation work at objective creation' do
    let!(:linked_lessor) { create(:lessor) }

    let!(:basic_housing_offer1) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        typology: 'T2',
      )
    end

    let!(:status1) do
      create(
        :status,
        object_id: basic_housing_offer1.id.to_s,
        status_label: 'lessor_status',
        title: 'PLACE'
      )
    end

    let!(:basic_housing_offer2) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        typology: 'T2',
      )
    end

    let!(:status2) do
      create(
        :status,
        object_id: basic_housing_offer2.id.to_s,
        status_label: 'lessor_status',
        title: 'PLACE',
      )
    end

    let!(:linked_lessor_objective) do
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:T2 =>100},
      )
    end

    it 'count when offer placed' do
      expect(linked_lessor_objective.target_by_typologies).to(eq({:T2=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq([:T2]))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({:T2=>2}))
    end
  end

  describe 'Realized calculation work at objective creation within old env' do
    before :each do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
    end

    let!(:linked_lessor) { create(:lessor) }

    let!(:basic_housing_offer1) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'PLACE',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # TODO: à supprimer une fois les statuts mis à jour
      )
    end

    let!(:basic_housing_offer2) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'PLACE',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # TODO: à supprimer une fois les statuts mis à jour
      )
    end
    let!(:basic_housing_offer3) do
      create(
          :housing_offer,
          lessor: linked_lessor,
          typology: 'T2',
          is_mad: true
      )
    end
    let!(:linked_lessor_objective) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:T2 =>100},
      )
    end

    it 'count when offer placed' do
      expect(linked_lessor_objective.target_by_typologies).to(eq({:T2=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq([:T2]))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({:T2=>2}))
    end

    it 'edit offer with nud trigger sopra flux' do
      after_save_res = basic_housing_offer3.after_save_actions
      #expect(after_save_res[:exectuted]).to eq(true)
    end
  end

  describe 'with large quantity' do
    nb = 150
    nb.times do |i|
      let!("offer#{i}") { create(:housing_offer) }
    end

    describe 'als_id is set' do
      let(:all_offers) do
        Hermes::HousingOffer.only(:als_id)
                            .where(als_id: { :$ne => nil })
      end

      it "has #{nb} entities" do
        expect(all_offers.count).to be nb
      end

      let(:als_ids) { all_offers.pluck(:als_id).uniq.compact }

      it 'produces valid lenght' do
        expect(als_ids.map(&:length)).to all(be(18))
      end

      it 'has no duplicates' do
        expect(
          als_ids.map { |als_id| all_offers.where(als_id: als_id).count }
        ).to all(be(1))
      end
    end
  end

  describe 'on creation PUBLIE' do
    let!(:offer_1) do
      create(
        :housing_offer,
        typology: 'T2',
        rent_with_charges: 434.32 + 53.12 + 22.43,
        rent_amount: 434.32,
        rental_charges: 53.12,
        parking_charges: 22.43,
        surface: 34,
        offer_status: 'PUBLIE',
        threshold_type: 'PLA LM',
        department: '34',
        threshold_subtype: 'Autres régions'
      )
    end

    it 'eligible_query ' do
      expect(offer_1.reload.eligibility_query).not_to be_nil
    end
  end

  describe 'Expiration settings on save' do
    let!(:lessor) { create :lessor }
    let(:published_offer) { create :published_offer, lessor_id: lessor.id }
    let(:long_publishing) { create :published_offer, lessor_id: lessor.id, expiry_days: 25 }

    shared_examples 'publication_end_date priority' do
      before do
        [published_offer, long_publishing].each do |offer|
          offer.update(publication_end_date: Time.new(2020, 3, 15, 0, 0, 0) + 20.days)
        end
      end
      it 'always sets the date from the field' do
        [published_offer, long_publishing].each do |offer|
          expect(offer.publication_end_date).to eq (Time.new(2020, 3, 15, 0, 0, 0) + 20.days)
        end
      end
    end

    context 'when no expiration setting for department' do
      it_behaves_like 'publication_end_date priority'

      it 'sets the minimal_publication_days' do
        expect(published_offer.minimal_publication_days).to eq 3
      end

      it 'computes the publication_end_date with default expiration' do
        expect(published_offer.publication_end_date.strftime('%d/%m/%Y')).to eq (Time.new(2020, 3, 15, 0, 0, 0) + 7.days).strftime('%d/%m/%Y')
      end

      it 'computes the publication_end_date with defined expiration' do
        expect(long_publishing.publication_end_date.strftime('%d/%m/%Y')).to eq (Time.new(2020, 3, 15, 0, 0, 0) + 25.days).strftime('%d/%m/%Y')
      end
    end

    context 'when expiration setting for department' do
      let!(:setting) { create :expiration_setting, expiry_days: 11 }

      it_behaves_like 'publication_end_date priority'

      it 'sets the minimal_publication_days' do
        expect(published_offer.minimal_publication_days).to eq 6
      end

      it 'sets the max_applicants' do
        expect(published_offer.max_applicants).to eq 10
      end

      it 'computes the publication_end_date with the setting expiration' do
        [published_offer, long_publishing].each do |offer|
          expect(offer.publication_end_date).to eq (Time.new(2020, 3, 15, 0, 0, 0) + 11.days)
        end
      end
    end

    context 'when expiration setting for department and lessor' do
      let!(:setting) { create :expiration_setting }
      let!(:setting_for_lessor) do
        create :expiration_setting, expiry_days: 8, minimal_publication_days: 4, max_applicants: 5, lessor_id: lessor.id
      end

      it_behaves_like 'publication_end_date priority'

      it 'sets the minimal_publication_days' do
        expect(published_offer.minimal_publication_days).to eq 4
      end

      it 'sets the max_applicants' do
        expect(published_offer.max_applicants).to eq 5
      end

      it 'computes the publication_end_date with the setting expiration' do
        [published_offer, long_publishing].each do |offer|
          expect(offer.publication_end_date).to eq(Time.new(2020, 3, 15, 0, 0, 0) + 8.days)
        end
      end
    end
  end

  describe 'Geocoding' do
    let!(:housing_offer) do
      create(
        :housing_offer,
        address: "vieux chemin d'arles",
        postal_code: '13210'
      )
    end
    it 'should call geocoder if address change' do
      enqueued_jobs.clear
      GeocodeJob.unlock_all
      housing_offer.address = 'routes des palus'
      housing_offer.save!
      housing_offer.reload

      job = enqueued_jobs.last
      expect(job[:job]).to eq(GeocodeJob)
      expect(job[:queue]).to eq('geocoder')
      expect(job[:args]).to eq(
        [
          housing_offer.class.name,
          housing_offer.id.to_s,
          true
        ]
      )
    end
  end

  describe 'threshold type and subtype' do
    let!(:threshold_plus) do
      create(
        :threshold,
        type: "PLUS",
        codes: ["PLU"],
        thresholds_map: {
          "Paris et communes limitrophes"=>{"thresholds_list"=>[23721, 35452, 46473, 55486, 66017, 74286], "another_person"=>8278},
          "IDF hors Paris et Communes limitrophes"=>{"thresholds_list"=>[23721, 35452, 42616, 51046, 60429, 68001], "another_person"=>7577},
          "Autres régions"=>{"thresholds_list"=>[20623, 27540, 33119, 39982, 47035, 53008], "another_person"=>5912}
        }
      )
    end

    let!(:threshold_plai) do
      create(
        :threshold,
        type: "PLA I",
        codes: ["PLN"],
        thresholds_map: {
          "Paris et communes limitrophes"=>{"thresholds_list"=>[13050, 21272, 27883, 30521, 36307, 40859], "another_person"=>4552},
          "IDF hors Paris et Communes limitrophes"=>{"thresholds_list"=>[13050, 21272, 25569, 28075, 33238, 37401], "another_person"=>4166},
          "Autres régions"=>{"thresholds_list"=>[11342, 16525, 19872, 22111, 25870, 29155], "another_person"=>3252}
        }
      )
    end

    let!(:offer_1) do
      create(
        :housing_offer,
        typology: 'T2',
        rent_with_charges: 434.32 + 53.12 + 22.43,
        rent_amount: 434.32,
        rental_charges: 53.12,
        parking_charges: 22.43,
        surface: 34,
        offer_status: 'PUBLIE',
        threshold_type: 'PLU',
        department: '34',
        threshold_subtype: 'Z3'
      )
    end

    it 'on creation' do
      expect(offer_1.reload.eligibility_query).not_to be_nil;
      expect(offer_1.reload.ceil_category_1).to eq(threshold_plus.thresholds_map["Autres régions"]["thresholds_list"][0]);
    end

    it 'threshold_type update' do
      offer_1.threshold_type = 'PLN'
      offer_1.save
      expect(offer_1.reload.eligibility_query).not_to be_nil;
      expect(offer_1.reload.ceil_category_1).to eq(threshold_plai.thresholds_map["Autres régions"]["thresholds_list"][0]);
    end

    it 'threshold_subtype update' do
      offer_1.threshold_subtype = 'Z1'
      offer_1.save
      expect(offer_1.reload.eligibility_query).not_to be_nil;
      expect(offer_1.reload.ceil_category_1).to eq(threshold_plus.thresholds_map["Paris et communes limitrophes"]["thresholds_list"][0]);
    end

    it 'threshold_type & threshold_subtype without mapping' do
      offer_1.threshold_type = 'PLA I'
      offer_1.threshold_subtype = 'IDF hors Paris et Communes limitrophes'
      offer_1.save
      expect(offer_1.reload.eligibility_query).not_to be_nil;
      expect(offer_1.reload.ceil_category_1).to eq(threshold_plai.thresholds_map["IDF hors Paris et Communes limitrophes"]["thresholds_list"][0]);
    end
  end

  describe '#handle_unpublished_status' do
    let(:published_offer) { create :published_offer, :with_applications }
    let(:request) { create :active_request }
    let(:request2) { create :active_request }
    let(:request3) { create :active_request }
    let(:new_application) { create :housing_application, housing_request: request, housing_offer: published_offer, status: "EN ATTENTE" }

    context 'on save' do
      it 'unpublishes the offer when publication_end_date is reached' do
        Timecop.freeze(Time.local(2020, 3, 22, 0, 0))
        expect {
          drain_all_jobs
        }.to change { published_offer.reload.offer_status }.from('LOGEMENT PUBLIE').to('COMMERCIALISATION NON PUBLIEE')
        .and change { published_offer.published }.from(true).to false
        Timecop.return
      end
    end
    
    context 'when max applicants is reached' do
      # minimal expiration date = 15/03/2020 + 3 days (default)
      context 'before minimal publication days' do
        before { Timecop.freeze(Time.local(2020, 3, 17, 0, 0)) }
        after  { Timecop.return }
        
        it 'sets the publication end date to the minimal expiration date' do
          expect {
            new_application2 = Hermes::HousingApplication.create(housing_request: request2, housing_offer: published_offer, status: "EN ATTENTE")
            new_application3 = Hermes::HousingApplication.create(housing_request: request3, housing_offer: published_offer, status: "EN ATTENTE")
          }.to change { published_offer.reload.publication_end_date.strftime("%d/%m/%Y") }.from('22/03/2020').to('18/03/2020')
        end
      end
      
      context 'after minimal publication days' do
        before { Timecop.freeze(Time.local(2020, 3, 21, 0, 0)) }
        after  { Timecop.return }

        it 'unpublishes the offer' do
          expect {
            new_application2 = Hermes::HousingApplication.create(housing_request: request2, housing_offer: published_offer, status: "EN ATTENTE")
            new_application3 = Hermes::HousingApplication.create(housing_request: request3, housing_offer: published_offer, status: "EN ATTENTE")
          }.to change { published_offer.reload.publication_end_date }.from('22/03/2020'.to_time).to(Time.now)
            .and change { published_offer.reload.offer_status }.from('LOGEMENT PUBLIE').to('COMMERCIALISATION NON PUBLIEE')
        end
      end
    end

    describe 'Hooks' do
      let!(:offer) {create(:housing_offer)}
      context 'Effort rate auto compute'  do
        it 'should have default value' do
          expect(offer.reload.effort_rate).not_to be_nil
        end

        it 'should have default value dependant of typology' do
          offer.update(typology: 't1', effort_rate: nil)
          expect(offer.reload.effort_rate).to eq(HERMES_CONFIG["EFFORT_RATES"]['T1'])
         
          offer.update(typology: 't6', effort_rate: nil)
          expect(offer.reload.effort_rate).to eq(HERMES_CONFIG["EFFORT_RATES"]['T6'])
        end

        it 'should have default value with invalid typology' do
          default_min_effort_rate = offer.min_effort_rate 
          offer.update(typology: 'azerty', effort_rate: nil)
          expect(offer.reload.effort_rate).not_to be_nil
          expect(offer.reload.effort_rate).to eq(default_min_effort_rate)
        end

        it 'use custom effort_rate if passed' do
          offer.update(typology: 't1', effort_rate: 12.1)
          expect(offer.reload.effort_rate).to eq(12.1)
        end
        
        it 'use custom or auto when is passed' do
          offer.update(typology: 't3', effort_rate: nil)
          expect(offer.reload.effort_rate).to eq(HERMES_CONFIG["EFFORT_RATES"]['T3'])
          
          offer.update(effort_rate: 12.1)
          expect(offer.reload.effort_rate).to eq(12.1)
          
          offer.update(effort_rate: nil, typology: 't1')
          expect(offer.reload.effort_rate).to eq(HERMES_CONFIG["EFFORT_RATES"]['T1'])
        end
      end

      context 'Charges auto compute' do
        let!(:valid_offer) { create(:housing_offer, rent_with_charges: 1250, rental_charges: 50, rent_amount: nil) }
        let!(:invalid_offer) { create(:housing_offer, rent_with_charges: nil, rental_charges: nil, rent_amount: nil) }

        it 'should auto fill rent_amount if blank' do
          expect(valid_offer.rent_amount).not_to be_nil
          expect(valid_offer.rent_amount).to eq(1250.0)
        end

        it 'should not auto fill rent_amount if cannot' do
          expect(invalid_offer.rent_amount).to be_nil
        end
      end
    end
  end

  describe 'lessor rights' do
    let(:lessor) { create :lessor, siren: '123456789' }
    let(:user) { create :user }
    let(:group_user) { create :user, group_ids:[lessor.group_id] }
    let(:supervisor_user) { create :user, group_ids:[lessor.supervisor_group_id] }
    let(:manager_user) { create :user, group_ids:[lessor.manager_group_id] }
    let(:offer) { create :housing_offer }

    it 'not add right without lessor' do
      offer.reload

      expect(Hermes::HousingOffer.find(offer.id).readable_by?(user)).to eq(false)
      expect(Hermes::HousingOffer.find(offer.id).readable_by?(group_user)).to eq(false)
      expect(Hermes::HousingOffer.find(offer.id).readable_by?(supervisor_user)).to eq(false)
      expect(Hermes::HousingOffer.find(offer.id).readable_by?(manager_user)).to eq(false)
    end

    it 'add right with lessor' do
      offer.lessor_id = lessor.id
      offer.save
      offer.reload
      expect(Hermes::HousingOffer.find(offer.id).readable_by?(user)).to eq(false)
      expect(Hermes::HousingOffer.find(offer.id).readable_by?(group_user)).to eq(true)
      expect(Hermes::HousingOffer.find(offer.id).writable_by?(supervisor_user)).to eq(true)
      expect(Hermes::HousingOffer.find(offer.id).writable_by?(manager_user)).to eq(true)
    end
  end

  describe 'housing_channel_keys' do 
    let!(:offer1) { create(:housing_offer)}
    let!(:offer2) { create(:housing_offer, housing_channel_keys: ['dalo'])}
    let!(:offer3) { create(:housing_offer, housing_channel_keys: ['dc','dalo'])}
    let!(:offer3) { create(:housing_offer, housing_channel_keys: ['dalo', 'lol'])}

    it 'create with and without housing_channel_keys' do 
      expect(offer1.housing_channel_keys).to eq(['dc'])
      expect(offer2.housing_channel_keys).to eq(['dalo'])
      expect(offer3.housing_channel_keys).to eq(['dalo', 'lol'])
    end 

    it 'when delete housing_channel_keys' do 
      offer2.update(housing_channel_keys: [])
      expect(offer2.reload.housing_channel_keys).to eq(['dc'])
    end 

    it 'when add housing_channel_keys' do 
      offer1.update(housing_channel_keys: ['dalo'])
      expect(offer1.reload.housing_channel_keys).to eq(['dalo'])
    end 

  end 

  describe 'reservation_holder' do 
    let!(:offer1) { create(:housing_offer, reserving_siren: nil)}
    let!(:rh) { create(:reservation_holder, name: 'test', siren: ['123456789'])}

    it 'default' do 
      expect(offer1.reload.reservation_holder_id).to eq(nil)
      expect(offer1.reload.reserving_siren).to eq(nil)
      expect(offer1.reload.reserving_business_name).to eq(nil)
    end 

    it 'changing siren' do
      expect do
        offer1.update(reserving_siren: rh.siren.first)
      end.to change { offer1.reload.reservation_holder_id}.from(nil).to(rh.id)
      expect(offer1.reload.reserving_business_name).to eq(rh.name)
    end 

    it 'wrong siren' do
      expect do
        offer1.reserving_siren = "923456789"
      end.to change { offer1.valid? }.from(true).to(false)
    end 
  end

  describe 'T6P' do 
    let!(:offer) { create(:housing_offer, insee_code: '34000', typology: 'T6P', offer_status: 'LOGEMENT PUBLIE', threshold_type: 'PLUS', threshold_subtype: 'Paris et communes limitrophes')}
    let!(:hr) {create(:complete_request)}

    it 'works' do
      expect(offer.reload.is_eligible_request(hr)).to eq false
      expect(hr.reload.is_eligible_offer(offer)).to eq false

      hr.update(dependents: Array.new(6, { rel_type: 'E', birth_date: '2000-01-01'}))
      hr.validate_all_required_document
      expect(offer.reload.is_eligible_request(hr)).to eq true
      expect(hr.reload.is_eligible_offer(offer)).to eq true

    end 
  end 

  describe 'corsica_exeption' do 
    let!(:offer) { create(:housing_offer, insee_code: '20000', typology: 'T1', offer_status: 'LOGEMENT PUBLIE', threshold_type: 'PLUS', threshold_subtype: 'Paris et communes limitrophes')}
    let!(:hr) { create(:complete_request, seeked_departments: ["2A", "22"], seeked_district_insee_codes: [])}
    let!(:hr2) { create(:complete_request, seeked_departments: ["2A", "22"], seeked_district_insee_codes: [])}


    it 'works' do
      expect(offer.reload.eligibility_query[:truly_seeked_departments]["$in"]).to eq(["2B","2A","20"])
      expect(hr.reload.eligibility_query[:department]["$in"]).to eq(["20","22","2A","2B"])
      expect(offer.reload.is_eligible_request(hr)).to eq true
      expect(hr.reload.is_eligible_offer(offer)).to eq true
    end 
  end 


  describe 'create_new_offers_notifications' do
    let!(:offer1) {create(:published_offer, :with_eligible_requests, requests_count: 3, date_publication_start: DateTime.now - 2.hours)}
    let!(:offer2) {create(:published_offer, :with_eligible_requests, requests_count: 2, date_publication_start: DateTime.now() - 2.days)}

    before do
      create_mail_template_default()
    end 

    it 'works' do 
      offer1.eligibles_requests.first.set(active_application_id: 'test')
      expect do
        Hermes::HousingOffer.create_new_published_offers_notifications(DateTime.now - 1.days, DateTime.now)
      end.to change { enqueued_jobs.select{|job| job[:job] == AsyncNotifierJob }.length }.by 4
      last_job = execute_last_job(AsyncNotifierJob)

      expect(last_job[:args][0]["datas"]["offer_ids"]).to eq([offer1.id.to_s])
    end 
  end 

  describe 'update_department' do 
    let!(:offer) {create(:housing_offer, postal_code: '24210')}

    it 'works for postal_code' do 
      expect(offer.reload.department).to eq('24')
    end 

    it 'works for insee_code' do 
      offer.update(insee_code: '33120')
      expect(offer.reload.department).to eq('33')
    end 

    it 'works for 97+' do 
      offer.update(insee_code: '97500')
      expect(offer.reload.department).to eq('975')
    end 

    it 'works for 2B' do 
      offer.update(insee_code: '2B500')
      expect(offer.reload.department).to eq('2B')
    end 

  end 

  describe 'main_picture' do 
    let!(:offer) {create(:housing_offer)}
    let!(:blob) { File.open([Hermes::Engine.root, %w[spec xml_bouchon minion.jpg]].flatten.join('/')).read }
    (1..5).each do |i|
      let!("picture#{i}".to_sym) {create(:picture, housing_offer: offer, thumb240_id: ImageTools.store_preview(blob, 240, 240))}
    end 

    it 'works for picture' do 
      expect(offer.main_picture).to eq(picture1)
      expect(offer.picture_for_email).to eq(picture1.get_absolute_asset('thumb240'))
      expect(offer.main_picture(2)).to eq(picture3)
      expect(offer.picture_for_email(2)).to eq(picture3.get_absolute_asset('thumb240'))
    end 

    it 'works with picture_urls' do
      offer.update(
        ordered_picture_ids: [],
        picture_urls: [
          { 'thumb240': 'http://localhost/test1.jpg'},
          { 'thumb240': 'http://localhost/test2.jpg'},
          { 'thumb240': 'http://localhost/test3.jpg'},
        ]
      ) 
      expect(offer.main_picture[:thumb240]).to eq('http://localhost/test1.jpg')
      expect(offer.picture_for_email).to eq('http://localhost/test1.jpg')
      expect(offer.main_picture(1)[:thumb240]).to eq('http://localhost/test2.jpg')
      expect(offer.picture_for_email(1)).to eq('http://localhost/test2.jpg')
    end 

    it 'works when no pictures' do 
      offer.update( ordered_picture_ids: [], picture_urls: [])
      expect(offer.main_picture).to eq(nil)
      expect(offer.picture_for_email).to eq(nil)
    end 
  end
  
  describe 'handle_is_mad' do 
    let(:offer) {create(:housing_offer)}

    it 'when is not mad' do 
      offer.update(offer_status:'EN ATTENTE COMMERCIALISATION')
      expect(offer.reload.designation_date_end).to eq(nil)
    end 

    it 'update is mad when no delai_reponse' do 
      offer.update(offer_status:'EN ATTENTE COMMERCIALISATION', lessor_status:'EN ATTENTE PUBLICATION')
      expect(offer.reload.designation_date_end.to_date).to eq((DateTime.current + 30.days).to_date)
    end   

    it 'update is mad with delai_reponse' do 
      offer.update(offer_status:'EN ATTENTE COMMERCIALISATION', lessor_status:'EN ATTENTE PUBLICATION', delai_reponse: 15)
      expect(offer.reload.designation_date_end.to_date).to eq((DateTime.current + 15.days).to_date)
    end  

    it 'with mad_als_date and delai_reponse' do 
      offer.update(offer_status:'EN ATTENTE COMMERCIALISATION', lessor_status:'EN ATTENTE PUBLICATION', delai_reponse: 15, mad_als_date: DateTime.current - 10.days)
      expect(offer.reload.designation_date_end.to_date).to eq((DateTime.current + 5.days).to_date)
    end 
  end 

  describe 'eligible with auto_request_validation' do 
    let!(:entity) { create(:entity, siret: '12345678901234', auto_request_validation: true)}
    let!(:applicant) {create(:applicant, :with_eligible_offers, offers_count: 1, main_employer_siret: entity.siret,)}
    let!(:hr) { applicant.first_active_request}
    let!(:offer) { hr.eligibles_offers.first }
    let!(:entity_reservation) { create(:entity_reservation, entity: entity, housing_offer: offer, from: Date.today - 1.days, to: Date.today + 1.days )}

    it 'works' do 
      hr.set(request_status: 'COMPLETE')
      expect(hr.entity_id).to eq(entity.id)
      expect(hr.entity_validated).to eq(true)
      expect(offer.reserved).to eq(true)


      expect(hr.is_eligible_offer(offer)).to eq(true)
      expect(offer.is_eligible_request(hr)).to eq(true)
    end 
  end 

  describe 'flux 13 & flux 11' do 
    let!(:user) {create(:user)}
    let!(:hr) { create(:complete_request, user_id: user.id)}
    let!(:offer) { create(:housing_offer)}
    let!(:ha) { create(:housing_application, status: 'EN ATTENTE', housing_request: hr, housing_offer: offer)}

    [
      {offer_status: 'RETIRE PAR LE BAILLEUR', lessor_status: 'RETIRE', ha_status: 'REPRIS BAILLEUR'},
      {offer_status: 'RETIRE PAR ALS', lessor_status: 'RETIRE', ha_status: 'REPRIS BAILLEUR'},
      {offer_status: 'ARCHIVE', lessor_status: 'RETIRE', ha_status: 'NON RETENUE'},
      {offer_status: 'RENDU AU BAILLEUR', lessor_status: 'RENDU AU BAILLEUR CONFIRME', ha_status: 'ANNULEE BAILLEUR'},
    ].each do |data|
      describe "when offer status #{data[:offer_status]} housing_application status #{data[:ha_status]}" do 
        it 'works' do
          offer.update(offer_status: data[:offer_status], lessor_status: data[:lessor_status])
          expect(enqueued_jobs.select {|job| job[:job] == Flux13SideEffectsJob || job[:job] == Flux11SideEffectsJob }.count).to eq(1)
          
          last_job = enqueued_jobs.select {|job| job[:job] == Flux13SideEffectsJob || job[:job] == Flux11SideEffectsJob}.last
          last_job[:job].perform_now(*deserialize_argument(last_job[:args]))
          expect(ha.reload.status).to eq(data[:ha_status])

          offer.save # check only send flux 13 when offer_status changed?
          expect(enqueued_jobs.select {|job| job[:job] == Flux13SideEffectsJob || job[:job] == Flux11SideEffectsJob}.count).to eq(1)
        end
      end
    end
  end 

  describe 'eligibility_query types' do 
    let!(:bt1) { create(:bordering_town, insee_code: '34200', bordering_towns: ['34210'])}
    let!(:bt2) { create(:bordering_town, insee_code: '34100', bordering_towns: ['34110', '34111'])}
    let!(:offer) { create(:published_offer, insee_code: '34210', typology: 'T1', threshold_type: 'PLUS', threshold_subtype: 'Paris et communes limitrophes') }
    let!(:hr_in_departement_1) { create(:complete_request, seeked_district_insee_codes: ['34000'])}
    let!(:hr_seeked) { create(:complete_request, seeked_district_insee_codes: ['34210', '34100'])}
    let!(:h_bordering) { create(:complete_request, seeked_district_insee_codes: ['34200', '34100'])}
    let!(:hr_in_departement_2) { create(:complete_request, seeked_district_insee_codes: ['34300'])}
    let!(:hr_out_of_departement) { create(:complete_request, seeked_district_insee_codes: ['99000'], seeked_departments: [])}

    it 'works' do
      # department
      expect(offer.eligibility_query[:truly_seeked_departments]).to eq('34')
      expect(offer.eligibles_requests.count).to eq(4)

      # seeked
      expect(offer.eligibility_query('seeked')[:seeked_district_insee_codes]).to eq('34210')
      expect(offer.eligibles_requests('seeked').count).to eq(1)
      expect(offer.is_eligible_request(hr_seeked, 'seeked')).to eq(true)

      # bordering
      expect(offer.eligibility_query('bordering')[:bordering_insee_codes]).to eq('34210')
      expect(offer.eligibles_requests('bordering').count).to eq(1)
      expect(offer.is_eligible_request(h_bordering, 'bordering')).to eq(true)

      # seeked_bordering
      expect(offer.eligibility_query('seeked_bordering')[:all_eligibility_codes]).to eq('34210')
      expect(offer.eligibles_requests('seeked_bordering').count).to eq(2)
      expect(offer.is_eligible_request(hr_seeked, 'seeked_bordering')).to eq(true)
      expect(offer.is_eligible_request(h_bordering, 'seeked_bordering')).to eq(true)
    
      # other_from_department
      expect(offer.eligibility_query('other_from_department')[:truly_seeked_departments]).to eq('34')
      expect(offer.eligibility_query('other_from_department')[:seeked_district_insee_codes]["$ne"]).to eq('34210')
      expect(offer.eligibility_query('other_from_department')[:bordering_insee_codes]["$ne"]).to eq('34210')
      expect(offer.eligibles_requests('other_from_department').count).to eq(2)
      expect(offer.is_eligible_request(hr_in_departement_1, 'other_from_department')).to eq(true)
      expect(offer.is_eligible_request(hr_in_departement_2, 'other_from_department')).to eq(true)

      # all_eligibility_query
      expect(Hermes::HousingRequest.where(offer.all_eligibility_query).count).to eq(4)
    end 

    it 'when missing data' do 
      offer.update(insee_code: nil, postal_code: nil)
      offer.reload
      expect(offer.eligibles_requests('department').count).to eq(0)
      expect(offer.eligibles_requests('seeked').count).to eq(0)
      expect(offer.eligibles_requests('seeked_bordering').count).to eq(0)
      expect(offer.eligibles_requests('other_from_department').count).to eq(0)
      expect(Hermes::HousingRequest.where(offer.all_eligibility_query).count).to eq(0)
    end 

    it 'all_eligibility_query corsica' do 
      expect(offer.all_eligibility_query[:all_eligibility_codes]).to eq({:$in => ['34', '34210']})
      offer.update(insee_code: '2A271')
      expect(offer.all_eligibility_query[:all_eligibility_codes]).to eq({:$in => ['2B', '2A', '20', '2A271']})
      expect(Hermes::HousingRequest.where(offer.all_eligibility_query).count).to eq(0)
    end 
  end 

  describe 'indexes eligibility' do 
    let!(:offer) { create(:housing_offer, typology: 'T1')}
    
    it 'works' do
      Hermes::HousingRequest.remove_indexes
      Hermes::HousingRequest.create_indexes
  
      expect(offer.eligibles_requests.explain.to_s).not_to include("COL")
      expect(offer.eligibles_requests.explain['queryPlanner']['winningPlan']['inputStage']['indexName']).to eq('department')
    end 
  end
  
  describe 'township_attachment eligibility' do 
    let!(:offer) { create(:housing_offer, :with_eligible_requests, requests_count: 3, insee_code: '34000' )}
    
    it 'recalculate eligibility' do 
      Hermes::HousingRequest.each_with_index do |hr, i|
        case i
        when 0
          hr.set(applicants_attached_codes: ['34000', '24000'])
        when 1
          hr.set(applicants_attached_codes: ['34000'])
        when 2
          hr.set(applicants_attached_codes: ['26000'])
        end
      end 

      expect(offer.eligibility_query[:applicants_attached_codes].present?).to eq(false)
      expect(offer.eligibles_requests.count).to eq(3)
      offer.update(housing_channel_keys: ['township_attachment'])
      expect(offer.reload.eligibility_query[:applicants_attached_codes]).to eq('34000')
      expect(offer.eligibles_requests.count).to eq(2)
    end 
  end

  describe 'update_has_disability_access' do 
    let!(:offer) {create(:housing_offer)}

    it 'works' do 
      expect(offer.disability_accesses).to eq(nil)
      expect(offer.has_disability_access).to eq(false)

      offer.update(disability_accesses: 3)
      expect(offer.has_disability_access).to eq(true)

      offer.update(disability_accesses: 2)
      expect(offer.has_disability_access).to eq(false)
    end 

  end 

  describe 'update_lessor_rights' do 
    let!(:lessor1) {create(:lessor, siren: '111111111')}
    let!(:lessor2) {create(:lessor, siren: '222222222')}
    let!(:offer) {create(:housing_offer, lessor_siren: lessor1.siren)}


    it 'works' do 
      expect(offer.reload.lessor_id).to eq(lessor1.id)
      expect(offer.reload.readers.to_s).to include(lessor1.group_id)
      expect(offer.reload.writers.to_s).to include(lessor1.supervisor_group_id)
      expect(offer.reload.writers.to_s).to include(lessor1.manager_group_id)
    end 

    it 'when update lessor' do 
      offer.update(lessor_siren: lessor2.siren)
      expect(offer.reload.lessor_id).to eq(lessor2.id)
      expect(offer.reload.readers.to_s).not_to include(lessor1.group_id)
      expect(offer.reload.writers.to_s).not_to include(lessor1.supervisor_group_id)
      expect(offer.reload.writers.to_s).not_to include(lessor1.manager_group_id)
      expect(offer.reload.readers.to_s).to include(lessor2.group_id)
      expect(offer.reload.writers.to_s).to include(lessor2.supervisor_group_id)
      expect(offer.reload.writers.to_s).to include(lessor2.manager_group_id)
    end 

  end 

  describe 'construction_year' do 
    let!(:offer) {create(:housing_offer)}

    it 'works with date' do 
      offer.update(construction_year: Date.today)
      expect(offer.construction_year).to eq(Date.today)
    end 

    it 'works with string year' do 
      offer.update(construction_year: "2017")
      expect(offer.construction_year).to eq(Date.new(2017))
    end 

    it 'works with date string' do 
      offer.update(construction_year: "2017-02-03")
      expect(offer.construction_year).to eq(Date.new(2017,2,3))
    end 

    it 'works when nil' do 
      offer.update(construction_year: Date.today)
      offer.update(construction_year: nil)
      expect(offer.construction_year).to be_nil
    end 

    it 'works when empty string' do 
      offer.update(construction_year: Date.today)
      offer.update(construction_year: '')
      expect(offer.construction_year).to be_nil
    end 

  end 

  describe 'calculate_eligibility_count_offer' do 
    let!(:offer) {create(:housing_offer, :with_eligible_requests, requests_count: 6)}
    let!(:old_eco) {create(:eligible_count_offer, housing_offer: offer, created_at: 8.day.ago)}
    let!(:entity1) {create(:entity)}
    let!(:reservation1) {create(:entity_reservation, housing_offer: offer, entity_id: entity1.id, from: Date.today - 1.days, to: Date.today + 1.days)}

    it 'works' do
      expect {
        Hermes::HousingOffer.calculate_eligibility_count_offer
      }.to change { Hermes::EligibleCountOffer.first.created_at > 1.days.ago}.from(false).to(true)
    end
  end 
end
